/*************************************************************************
 * BTSMS
 * BTSMS is an application that provides support to connect bluetooth enabled
 * mobile phones from a PC and use them to send SMS. The SMSes can be written
 * over PC using the comfort of the keyboard rather than using small keypad
 * of the mobile phone
 *
 * Official Website: http://bt-tools.sourceforge.net
 * -------------------------------------------------------------------------
 * 
 * Name    : btlib.c
 * 
 * Purpose : Source file for BTLIB (Bluetooth wrapper library over BlueZ)
 *
 * Authors : Shobhit Arora
 * 
 * License : GNU Public License (Version 2 or later)
 *
 * -------------------------------------------------------------------------
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' BY THE AUTHORS AND INDEPENDENT
 * CONTRIBUTORS AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 **************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "btlib.h"

BT_DEV_LIST bt_dev_list;
#if 0
BT_SRV_INFO     service_name_array[BT_NUM_SRV] = {
    { "0x1000", "SDP Server" },
    { "0x1103", "Dial-up Networking" },
    { "0x1112", "Headset Audio gateway" },
    { "0x111f", "Handfree Audio Gateway" },
    { "0x1105", "OBEX Object Push", },
    { "0x1106", "OBEX File Transfer" },
    { "0x111b", "Imaging Responder" },
    { "0x110a", "Audio Source" },
    { "0x110c", "AV Remote Target" },
    { "0x1203", "Generic Audio" }
};
#endif
/* Function to discover bluetooth devices */
INT16 btlib_dev_find(BT_DEV_LIST_PTR dev_list, UINT16 *dev_count, UINT16 *error)
{
    inquiry_info *inq_info = NULL;
    INT32 max_dev, num_dev;
    INT32 dev_id, sock, len, flags;
    INT32 i;
    INT8 name[BDNAME_LEN] = { 0 };

    *error = BTLIB_ERR_OK;

    dev_id = hci_get_route(NULL);
    sock = hci_open_dev(dev_id);
    if(dev_id < 0 || sock , 0)
    {
        *error = BTLIB_ERR_SOCK;
        *dev_count = 0;
        return BTLIB_ERR_FAIL;
    }        

    len = INQ_LEN;
    max_dev = NUM_DEV_MAX;

    flags = IREQ_CACHE_FLUSH;
    inq_info = (inquiry_info *)malloc(max_dev * sizeof(inquiry_info));
    
    /* search for devices */
    num_dev = hci_inquiry(dev_id, len, max_dev, NULL, &inq_info, flags);
    if(!num_dev)
    {
        *error = BTLIB_ERR_NO_DEV;
        *dev_count = 0;
        return BTLIB_ERR_OK;
    }

    *dev_count = num_dev;

    /* Find the device names */
    for(i = 0; i < num_dev; i++)
    {
        memset(name, 0, sizeof(name));
        if(hci_read_remote_name(sock, &(inq_info + i)->bdaddr, sizeof(name), name, 0) < 0)
            strcpy(name, "[Unknown]");

        memcpy(&dev_list->bt_dev_info[i].addr, &(inq_info + i)->bdaddr, 6);
        strcpy(dev_list->bt_dev_info[i].name, name);
        memcpy(dev_list->bt_dev_info[i].dev_class, (inq_info + i)->dev_class, 3);
    }

    dev_list->num_dev = i;
    free(inq_info);
    close(sock);
    return 0;
}

/* Function to display the device list */
void btlib_show_devices(BT_DEV_LIST *pdev_list)
{
    UINT16 i;
    INT8 addr[BDADDR_LEN] = { 0 };

    printf("\nNumber of devices = %d\n", pdev_list->num_dev);

    for(i = 0; i< pdev_list->num_dev; i++)
    {
        ba2str(&pdev_list->bt_dev_info[i].addr, addr);
        printf("%d. Name: %s <---> Address: %s\n", i+1, pdev_list->bt_dev_info[i].name, addr);
#if 0
        printf("Class: 0x%x %x %x\n", pdev_list->bt_dev_info[i].dev_class[0],
               pdev_list->bt_dev_info[i].dev_class[1], pdev_list->bt_dev_info[i].dev_class[2] );
#endif
    }
}

/* Function to search for the available services on a specified device */
int btlib_sdp_browse(bdaddr_t *dev_addr, sdp_list_t **psdp_list, UINT16 *error)
{
    sdp_session_t *session;
    sdp_list_t  *search, *attr, *list, *list_cpy;
    int retVal;
    uuid_t uuid;
    uint32_t range = 0x0000ffff;

    *error = BTLIB_ERR_OK;

    session = sdp_connect(BDADDR_ANY, dev_addr, 0);
    if(!session)
    {
        *psdp_list = NULL;
        *error = BTLIB_ERR_SDP_CONNECT;
        return BTLIB_ERR_FAIL;
    }

    sdp_uuid16_create(&uuid, PUBLIC_BROWSE_GROUP);
    search = sdp_list_append(0, &uuid);
    attr = sdp_list_append(0, &range);

    /* browse for a list of services and their attributes */
    retVal = sdp_service_search_attr_req(session, search, SDP_ATTR_REQ_RANGE, attr, psdp_list);
    if(retVal < 0)
    {
        printf("sdp search failed\n");
        *psdp_list = NULL;
        *error = BTLIB_ERR_SDP_FIND;
        return BTLIB_ERR_FAIL;
    }

    sdp_close(session);
    return BTLIB_ERR_OK;
}

int btlib_sdp_browse_all(BT_DEV_LIST_PTR *ppdev_list, UINT16 *error)
{
    INT16 i, retVal;
    BT_DEV_LIST_PTR plist = *ppdev_list;
    BT_DEV_INFO_PTR pdev_info;
    *error = BTLIB_ERR_OK;

    for(i = 0; i < plist->num_dev; i++)
    {
        pdev_info = &(plist->bt_dev_info[i]);
        retVal = btlib_sdp_browse(&(pdev_info->addr), &(pdev_info->service_list), error); 
        if(BTLIB_ERR_OK != retVal)
        {
            if(BTLIB_ERR_SDP_CONNECT == *error)
            {
                pdev_info->connected = 0;
            }
            return retVal;
        }
        pdev_info->connected = 1;
    }

    return BTLIB_ERR_OK;
}


INT8 uuidstr[MAX_LEN_UUID_STR];
INT8 ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
INT8 str[MAX_LEN_PROFILEDESCRIPTOR_UUID_STR];

void btlib_sdp_info_show(BT_DEV_LIST_PTR pdev_list, INT16 dev_index)
{
    BT_DEV_INFO_PTR pdev_info;
    INT16 i;
    sdp_profile_desc_t *desc = 0;
    sdp_list_t *srv_list, *protos, *profile_desc =  0;

    INT8 srv_name[64];
    INT32 rfcomm_channel;
    INT8 addr[BDADDR_LEN] = { 0 };

    if(dev_index > pdev_list->num_dev || dev_index <=0)
    {
        printf("No such device !!\n");
        return;
    }

    pdev_info = &pdev_list->bt_dev_info[dev_index - 1];
    srv_list = pdev_info->service_list;
    ba2str(&(pdev_info->addr), addr);
    printf("%s --> %s\n", pdev_info->name, addr);
    while(srv_list != NULL)
    {
        sdp_get_profile_descs((sdp_record_t *)srv_list->data, &profile_desc);
        sdp_get_access_protos((sdp_record_t *)srv_list->data, &protos);
        rfcomm_channel = sdp_get_proto_port(protos, PROTO_RFCOMM);
        
        if(!profile_desc)
        {
            srv_list = srv_list->next;
            continue;
        }

        desc = (sdp_profile_desc_t *)profile_desc->data;

        sdp_profile_uuid2strn(&desc->uuid, uuidstr, MAX_LEN_UUID_STR);
        if(rfcomm_channel)
        {
            if(strlen(uuidstr) < 15)
                printf("\tProfile: %s\t\tChannel:%d\n", uuidstr, rfcomm_channel);
            else
                printf("\tProfile: %s\tChannel:%d\n", uuidstr, rfcomm_channel);
        }
        else
        {
            if(strlen(uuidstr) < 15)
                printf("\tProfile: %s\t\tRFCOMM: None\n", uuidstr);
            else
                printf("\tProfile: %s\t\tRFCOMM: None\n", uuidstr);
        }

        srv_list = srv_list->next;
    }
}

INT16 btlib_get_dialup_rfcomm_port(BT_DEV_INFO_PTR pDev)
{
    INT8 srv_name[64];
    sdp_list_t *srv_list, *protos, *profile_desc;
    sdp_profile_desc_t *desc = 0;
    srv_list = pDev->service_list;

    while(srv_list != NULL)
    {
        sdp_get_profile_descs((sdp_record_t *)srv_list->data, &profile_desc);
        if(!profile_desc)
        {
            srv_list = srv_list->next;
            continue;
        }

        desc = (sdp_profile_desc_t *)profile_desc->data;

        sdp_profile_uuid2strn(&desc->uuid, uuidstr, MAX_LEN_UUID_STR);

        if(strstr(uuidstr, "Dial"))
        {
            sdp_get_access_protos((sdp_record_t *)srv_list->data, &protos);
            return (sdp_get_proto_port(protos, PROTO_RFCOMM));
        }

        srv_list = srv_list->next;
    }

    return BTLIB_ERR_FAIL;
}

INT16 btlib_get_obex_channel(BT_DEV_INFO_PTR pDev)
{
    INT8 srv_name[64];
    sdp_list_t *srv_list, *protos, *profile_desc;
    sdp_profile_desc_t *desc = 0;
    srv_list = pDev->service_list;

    while(srv_list != NULL)
    {
        sdp_get_profile_descs((sdp_record_t *)srv_list->data, &profile_desc);
        if(!profile_desc)
        {
            srv_list = srv_list->next;
            continue;
        }

        desc = (sdp_profile_desc_t *)profile_desc->data;

        sdp_profile_uuid2strn(&desc->uuid, uuidstr, MAX_LEN_UUID_STR);

        if(strstr(uuidstr, "OBEX Object Push"))
        {
            sdp_get_access_protos((sdp_record_t *)srv_list->data, &protos);
            return (sdp_get_proto_port(protos, PROTO_RFCOMM));
        }

        srv_list = srv_list->next;
    }

    return BTLIB_ERR_FAIL;
}


int btlib_rfcomm_connect(bdaddr_t *paddr, UINT16 channel, INT32 *socket_fd, UINT16 *error)
{
    struct sockaddr_rc saddr = { 0 };    
    UINT32 status;

    *error = BTLIB_ERR_OK;

    *socket_fd = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
    if(*socket_fd < 0)
    {
        *error = BTLIB_ERR_SOCK;
        return BTLIB_ERR_FAIL;
    }

    /* set the connection parameters */
    saddr.rc_family = AF_BLUETOOTH;
    saddr.rc_channel = channel;
    memcpy(&(saddr.rc_bdaddr), paddr, sizeof(bdaddr_t));

    /* connect to the server */
    status = connect(*socket_fd, (struct sockaddr *)&saddr, sizeof(saddr));
    if(status < 0)
    {
        *error = BTLIB_ERR_RFCOMM_CON;
        return BTLIB_ERR_FAIL;
    }

    return BTLIB_ERR_OK;
}

void btlib_rfcomm_close(INT32 socket)
{
    close(socket);
}

int btlib_rfcomm_write(INT32 socketfd, INT8 *data, UINT32 len)
{
    UINT32 status;
    status = write(socketfd, data, len);
    if(status < 0)
    {
        return BTLIB_ERR_FAIL;
    }

    return status;
}

int btlib_rfcomm_read(INT32 socketfd, INT8 *data, UINT32 len)
{
    UINT32 status;
    status = read(socketfd, data, len);
    if(status < 0)
    {
        return BTLIB_ERR_FAIL;
    }

    return status;
}
